package com.jeffknecht.dynadao.api;

import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.HashMap;
import java.util.List;

import com.jeffknecht.dynadao.annotation.DTOClass;
import com.jeffknecht.dynadao.annotation.GroupEnd;
import com.jeffknecht.dynadao.annotation.GroupStart;
import com.jeffknecht.dynadao.annotation.Or;
import com.jeffknecht.dynadao.annotation.OrderBy;
import com.jeffknecht.dynadao.annotation.Parameter;
import com.jeffknecht.dynadao.annotation.Parameter.Operator;
import com.jeffknecht.dynadao.annotation.Where;

/**
 * <p>
 * Provides a dynamic proxy for a DAO interface.
 * </p>
 * 
 * <p>
 * There are some specific rules the DAO interface must follow:
 * </p>
 * 
 * <p>First, the interface may only declare the following methods:
 * <ul>
 * <li>int insert(Object<T>)</li>
 * <li>int update(Object<T>)</li>
 * <li>int delete(Object<T>)</li>
 * <li>int countAll()</li>
 * <li>int countBy...(args...)</li>
 * <li>List<T> findAll()</li>
 * <li>T findBy...(args...)</li>
 * <li>List<T> findBy...(args...)</li>
 * </ul>
 * </p>
 * 
 * <p>Second, either the interface or each of the find methods
 * must be annotated with the DTOClass annotation, defining
 * the class (<T>) that is returned by the find... methods</p>
 * 
 * <p>Third, every parameter of a find... method must be
 * annotated with a Parameter annotation, defining the property
 * in the DTOClass that is represented by the parameter</p>
 * 
 * <p>Any call to a method that does not follow these guidelines
 * may be met with a nasty runtime exception. You have been warned!</p>
 * 
 * @author jeff.knecht
 *
 */
public class DynaDaoProxy {

	/**
	 * Builds and returns a DynaDao proxy based on the supplied interface.
	 * 
	 * @param intfce The interface to proxy
	 * @param dynaDao The DynaDao instance to pass the calls to.
	 * @return
	 */
	public static <T> T create(Class<T> intfce, DynaDao dynaDao) {
		InvocationHandler handler = new DynaDaoProxyInvocationHandler(intfce, dynaDao);
		return (T) Proxy.newProxyInstance(Thread.currentThread().getContextClassLoader(), new Class<?>[] {intfce}, handler);
	}
	
	private static final class DynaDaoProxyInvocationHandler implements InvocationHandler {
		private DynaDao dynaDao;
		private Class<?> proxyInterface;
		
		private DynaDaoProxyInvocationHandler(Class proxyInterface, DynaDao dynaDao) {
			this.dynaDao = dynaDao;
			this.proxyInterface = proxyInterface;
		}
		
		public Object invoke(Object proxy, Method method, Object[] args)
				throws Throwable {
			String methodName = method.getName();
			
			if (methodName.startsWith("insert")) {
				return dynaDao.insert(args[0]);
			} else if (methodName.startsWith("update")) {
				return dynaDao.update(args[0]);
			} else if (methodName.startsWith("delete")) {
				return dynaDao.delete(args[0]);
			} else if (methodName.equals("countAll")) {
				DTOClass type = getDTOClass(method);
				return dynaDao.countAll(type.value());
			} else if (methodName.startsWith("count")) {
				DTOClass type = getDTOClass(method);
				
				HashMap<String, Object> params = new HashMap<String, Object>();
				StringBuilder sb = new StringBuilder();
				
				buildDynaDaoWhereClause(method, args, params, sb);
			
				return dynaDao.count(type.value(), sb.toString(), params);
			} else if (methodName.equals("findAll")) {
				DTOClass type = getDTOClass(method);
				OrderBy orderBy = method.getAnnotation(OrderBy.class);
				if (orderBy != null) {
					return dynaDao.findAll(type.value(), buildDynaDaoOrderByClause(orderBy.value()));					
				} else {
					return dynaDao.findAll(type.value());
				}
			} else if (methodName.startsWith("find")) {
				DTOClass type = getDTOClass(method);
				
				HashMap<String, Object> params = new HashMap<String, Object>();
				StringBuilder sb = new StringBuilder();
				
				buildDynaDaoWhereClause(method, args, params, sb);
			
				if (method.getReturnType().isAssignableFrom(List.class)) {
					OrderBy orderBy = method.getAnnotation(OrderBy.class);
					if (orderBy != null) {
						return dynaDao.find(type.value(), sb.toString(), params, buildDynaDaoOrderByClause(orderBy.value()));
					} else {
						return dynaDao.find(type.value(), sb.toString(), params);
					}
				} else {
					return dynaDao.findOne(type.value(), sb.toString(), params);
				}
			} else {
				throw new UnsupportedOperationException("Unsupported method: " + methodName);
			}
		}

		private String buildDynaDaoOrderByClause(String[] attributes) {
			StringBuilder sb = new StringBuilder();
			for (String attr : attributes) {
				if (sb.length() > 0) {
					sb.append(", ");
				}
				sb.append("{");
				String[] parts = attr.split("\\s+");
				for (int i = 0; i < parts.length; i++) {
					if (i > 0) {
						sb.append(" ");
					}
					sb.append(parts[i]);
					if (i == 0) {
						sb.append("}");
					}
				}
			}
			return sb.toString();
		}
		
		/**
		 * 
		 * @param method The method that is being called
		 * @param args The argument values for the method
		 * @param params A Map to hold the generated parameter values
		 * @param sb A StringBuilder to append the where clause to
		 */
		private void buildDynaDaoWhereClause(Method method, Object[] args,
				HashMap<String, Object> params, StringBuilder sb) {
			
			StringBuilder staticWhereBuilder = new StringBuilder();
			Where whereAnnotation = method.getAnnotation(Where.class);
			if (whereAnnotation != null) {
				staticWhereBuilder.append(whereAnnotation.value().trim());
			}
			
			StringBuilder dynamicWhereBuilder = new StringBuilder();
			Annotation[][] paramAnnotations = method.getParameterAnnotations();
			int openParens = 0;
			for (int i = 0; i < paramAnnotations.length; i++) {
				boolean newParam = true;
				for (int j = 0; j < paramAnnotations[i].length; j++) {
					Annotation paramAnnot = paramAnnotations[i][j];
					if (newParam) {
						if (paramAnnot instanceof Or) {
							if (dynamicWhereBuilder.length() > 0) {
								dynamicWhereBuilder.append(" or ");
							}
						} else {
							if (dynamicWhereBuilder.length() > 0) {
								dynamicWhereBuilder.append(" and ");
							}
						}
						newParam = false;
					}
					

					if (paramAnnot instanceof GroupStart) {
						openParens++;
						dynamicWhereBuilder.append("(");
					}
					
					if (paramAnnot instanceof GroupEnd) {
						openParens--;
						dynamicWhereBuilder.append(")");
					}
					
					if (paramAnnot instanceof Parameter) {
						String paramName = ((Parameter) paramAnnot).value();
						String placeHolder = ((Parameter) paramAnnot).placeholder();
						Operator comparison = ((Parameter) paramAnnot).comparison();
						boolean isComparisonIn = comparison.equals(Operator.IN);
						if (placeHolder.trim().equals("")) {
							placeHolder = paramName;
						}
						dynamicWhereBuilder.append("{" + paramName + "} " + comparison.sql() + (isComparisonIn ? "(" : " ") + ":" + placeHolder + (isComparisonIn ? ")" : ""));
						params.put(placeHolder, args[i]);
					}
				}
			}
			while (openParens > 0) {
				dynamicWhereBuilder.append(")");
				openParens--;
			}
			
			StringBuilder finalBuilder = new StringBuilder();
			if (staticWhereBuilder.length() > 0) {
				finalBuilder.append(staticWhereBuilder.toString());
				if (dynamicWhereBuilder.length() > 0) {
					finalBuilder.append(" and (").append(dynamicWhereBuilder.toString()).append(")");
				}
			} else {
				finalBuilder.append(dynamicWhereBuilder.toString());
			}
			sb.append(finalBuilder.toString());
		}

		private DTOClass getDTOClass(Method method) {
			DTOClass type = method.getAnnotation(DTOClass.class);
			if (type == null) {
				type = proxyInterface.getAnnotation(DTOClass.class);
			}
			return type;
		}
		
	}
	
}
